! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful,
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************

      module run_star_extras

      use star_lib
      use star_def
      use const_def
      use crlibm_lib
      use interp_2d_lib_sg

      implicit none

      ! #  Ranges of state parameters (AESOPUS)
      ! #
      ! #  log10(T) range: nt=  67  values from  4.500  to  3.200
      ! #  log10(T) = [3.200,3.700] in steps of  0.010
      ! #  log10(T) = ]3.700,4.500] in steps of  0.050
      ! #
      ! #  log10(R) range: nre= 19  values from -8.000  to  1.000
      ! #                           in steps of  0.500
      
      integer, parameter :: kk_nt = 67
      integer, parameter :: kk_nr = 19

      real, parameter, dimension(kk_nt) :: kk_t = &
           (/3.20, 3.21, 3.22, 3.23, 3.24, 3.25, 3.26, 3.27, 3.28, &
             3.29, 3.30, 3.31, 3.32, 3.33, 3.34, 3.35, 3.36, 3.37, &
             3.38, 3.39, 3.40, 3.41, 3.42, 3.43, 3.44, 3.45, 3.46, 3.47, &
             3.48, 3.49, 3.50, 3.51, 3.52, 3.53, 3.54, 3.55, 3.56, 3.57, &
             3.58, 3.59, 3.60, 3.61, 3.62, 3.63, 3.64, 3.65, 3.66, 3.67, &
             3.68, 3.69, 3.70, 3.75, 3.80, 3.85, 3.90, 3.95, 4.00, 4.05, &
             4.10, 4.15, 4.20, 4.25, 4.30, 4.35, 4.40, 4.45, 4.50/)

      real, parameter, dimension(kk_nr) :: kk_r = &
           (/-8.0, -7.5, -7.0, -6.5, -6.0, -5.5, -5.0, -4.5, -4.0, &
           -3.5, -3.0, -2.5, -2.0 , -1.5, -1.0 , -0.5, 0.0 , 0.5, 1.0/)

      ! FC = log10(variation factor) of carbon abundance
      ! mc =  5 values:    0.000   0.250   0.500   0.750   1.000
      ! mc =  4 values:    1.250   1.500   1.750   2.000
      integer, parameter :: kk_nc = 9
      real, parameter, dimension(kk_nc) :: kk_c = &
           (/0.00342, 0.00608, 0.0108, 0.0192, & 
           0.0342, 0.0608, 0.108, 0.192, &
           0.342/)

      real, dimension(:,:,:), allocatable :: kk_data

      ! these routines are called by the standard run_star check_model
      contains

      subroutine extras_controls(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         real :: log10_kap
         real :: logt
         real, dimension(kk_nr) :: row
         character :: first_char
         integer :: uin, ios
         integer :: idx_t, idx_r, idx_c
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return

         ! this is the place to set any procedure pointers you want to change
         ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)

         ! allocate the array to hold the opacity data
         allocate(kk_data(kk_nc,kk_nr,kk_nt))

         ! this read routine will make you cry
         idx_c = 1
         idx_t = kk_nt
         
         ! read AESOPUS opacities
         open(newunit = uin, file = "AESOPUS-dXC.dat")

         do 
            if (idx_t .eq. 0) then
               idx_t = kk_nt
               idx_c = idx_c + 1
            endif
            read(uin, '(a1)', advance='no', iostat=ios) first_char
            if (ios .ne. 0) exit
            if (first_char .eq. '#') then
               read(uin,*) ! go to next line
            else
               read(uin, *) logt, row
               kk_data(idx_c, 1:kk_nr, idx_t) = row
               idx_t = idx_t - 1
            endif
         end do

         close(uin)

         s% other_kap_get_Type2 => cold_CO_kap_get_Type2
         s% other_D_mix => reduced_D_mix

         ! Uncomment these lines if you wish to use the functions in this file,
         ! otherwise we use a null_ version which does nothing.
         s% extras_startup => extras_startup
         s% extras_check_model => extras_check_model
         s% extras_finish_step => extras_finish_step
         s% extras_after_evolve => extras_after_evolve
         s% how_many_extra_history_columns => how_many_extra_history_columns
         s% data_for_extra_history_columns => data_for_extra_history_columns
         s% how_many_extra_profile_columns => how_many_extra_profile_columns
         s% data_for_extra_profile_columns => data_for_extra_profile_columns

           

         ! Once you have set the function pointers you want,
         ! then uncomment this (or set it in your star_job inlist)
         ! to disable the printed warning message,
         s% job% warn_run_star_extras =.false.
         
      end subroutine extras_controls

      subroutine aesopus_C(idx_c, log10_rho, log10_T, kap, dln_kap_dlnRho, dln_kap_dlnT)
        integer, intent(in) :: idx_c
        real(dp), intent(in) :: log10_rho
        real(dp), intent(in) :: log10_T
        real(dp), intent(out) :: kap
        real(dp), intent(out) :: dln_kap_dlnRho
        real(dp), intent(out) :: dln_kap_dlnT

        integer :: ierr
        
        real :: logd, logt, logr

        integer, parameter :: npts = 5
        real, dimension(npts) :: dpts, tpts, rpts, log10_kap

        real, dimension(kk_nr, kk_nt) :: kk_k
        real, dimension(3,kk_nr,kk_nt) :: wk

        real :: f
        real, parameter :: eps = 0.01 ! for numerical derivatives

        logd = real(log10_Rho)
        logt = real(log10_T)
        logr = logd - 3 * logt + 18

        dpts = (/logd, logd, logd, logd-eps, logd+eps/)
        tpts = (/logt, logt-eps, logt+eps, logt, logt/)
        rpts = dpts - 3 * tpts + 18

        kk_k = kk_data(idx_c, :, :)

        call do_RGBI3P_sg(1,kk_nr,kk_nt,kk_r,kk_t,kk_k,&
             npts,rpts,tpts,log10_kap,ierr,wk)

        kap = 10d0 ** log10_kap(1)

        f = 2d0 * eps
        
        dln_kap_dlnRho = (log10_kap(5)-log10_kap(4))/f
        dln_kap_dlnT = (log10_kap(3)-log10_kap(2))/f

      end subroutine aesopus_C

      subroutine aesopus(XC, log10_rho, log10_T, kap, dln_kap_dlnRho, dln_kap_dlnT, ierr)
        real(dp), intent(in) :: XC
        real(dp), intent(in) :: log10_rho
        real(dp), intent(in) :: log10_T
        real(dp), intent(out) :: kap
        real(dp), intent(out) :: dln_kap_dlnRho
        real(dp), intent(out) :: dln_kap_dlnT
        integer, intent(out) :: ierr

        integer :: idx_c

        real(dp) :: alfa, beta
        real(dp) :: kap1, kap2
        real(dp) :: dln_kap_dlnRho1, dln_kap_dlnRho2
        real(dp) :: dln_kap_dlnT1, dln_kap_dlnT2

        real(dp) :: log10_R
        
        log10_R = log10_Rho - 3 * log10_T + 18
        
        ! raise error if off of table
        if ((log10_T .lt. kk_t(1)) .or. (log10_T .gt. kk_t(kk_nt)) .or. &
            (log10_R .lt. kk_r(1)) .or. (log10_R .gt. kk_r(kk_nr))) then
           ierr = -1
           return
        end if

        ! too low, call and return
        if (XC .lt. kk_c(1)) then
           call aesopus_C(1, log10_rho, log10_T,  &
                kap, dln_kap_dlnRho, dln_kap_dlnT)
           return
        endif

        ! too high, call and return
        if (XC .ge. kk_c(kk_nc)) then
           call aesopus_C(kk_nc, log10_rho, log10_T,  &
                kap, dln_kap_dlnRho, dln_kap_dlnT)
           return
        endif

        ! get index O(N), but N small :)
        do idx_c = 1, kk_nc - 1
           if ((XC .ge. kk_c(idx_c)) .and. (XC .lt. kk_c(idx_c+1))) exit
        enddo

        alfa = (kk_c(idx_c+1)-XC)/ (kk_c(idx_c+1) - kk_c(idx_c))
        beta = 1 - alfa

        call aesopus_C(idx_c, log10_rho, log10_T,  &
             kap1, dln_kap_dlnRho1, dln_kap_dlnT1)
        call aesopus_C(idx_c+1, log10_rho, log10_T,  &
             kap2, dln_kap_dlnRho2, dln_kap_dlnT2)

        kap = alfa * kap1 + beta * kap2
        dln_kap_dlnRho = (alfa * kap1 * dln_kap_dlnRho1 + beta * kap2 * dln_kap_dlnRho2)/kap
        dln_kap_dlnT = (alfa * kap1 * dln_kap_dlnT1 + beta * kap2 * dln_kap_dlnT2)/kap
        
        return
        
      end subroutine aesopus
      

      subroutine cold_CO_kap_get_Type2( &
           id, k, handle, zbar, X, Z, Zbase, XC, XN, XO, XNe, &
           log10_rho, log10_T, species, chem_id, net_iso, xa, &
           lnfree_e, d_lnfree_e_dlnRho, d_lnfree_e_dlnT, use_Zbase_for_Type1, &
           frac_Type2, kap, dln_kap_dlnRho, dln_kap_dlnT, ierr)

        use kap_lib, only: kap_get_Type2
        integer, intent(in) :: id
        integer, intent(in) :: k
        integer, intent(in) :: handle
        real(dp), intent(in) :: zbar
        real(dp), intent(in) :: X, Z, Zbase, XC, XN, XO, XNe
        real(dp), intent(in) :: log10_rho
        real(dp), intent(in) :: log10_T
        double precision, intent(in) :: lnfree_e, d_lnfree_e_dlnRho, d_lnfree_e_dlnT
        integer, intent(in) :: species
        logical, intent(in) :: use_Zbase_for_Type1
        integer, pointer :: chem_id(:)
        integer, pointer :: net_iso(:)
        real(dp), intent(in) :: xa(:)
        real(dp), intent(out) :: frac_Type2
        real(dp), intent(out) :: kap
        real(dp), intent(out) :: dln_kap_dlnRho
        real(dp), intent(out) :: dln_kap_dlnT
        integer, intent(out) :: ierr

        real(dp), parameter :: log10_T_cold = 4.1 !3.80
        real(dp), parameter :: log10_T_warm = 4.2 !3.88
        real(dp) :: w_kap, w_dln_kap_dlnRho, w_dln_kap_dlnT
        real(dp) :: c_kap, c_dln_kap_dlnRho, c_dln_kap_dlnT
        real(dp) :: alfa, beta, logT, lower_bdy, upper_bdy

        type (star_info), pointer :: s
        call get_star_ptr(id, s, ierr)

        ! aliases
        logT = log10_T
        lower_bdy = log10_T_cold
        upper_bdy = log10_T_warm

        alfa = (logT - lower_bdy) / (upper_bdy - lower_bdy)
        alfa = max(min(alfa,1d0),0d0) ! force in [0,1]
        beta = 1d0 - alfa
        
        if (logT >= lower_bdy) then
           call kap_get_Type2( &
                handle, zbar, X, Z, Zbase, XC, XN, XO, XNe, &
                log10_rho, log10_T, &
                lnfree_e, d_lnfree_e_dlnRho, d_lnfree_e_dlnT, use_Zbase_for_Type1, &
                frac_Type2, w_kap, w_dln_kap_dlnRho, w_dln_kap_dlnT, ierr)
        else
           w_kap = 0
           w_dln_kap_dlnRho = 0
           w_dln_kap_dlnT = 0
        end if

        if (logT <= upper_bdy) then
           call aesopus(XC, log10_rho, log10_T,  &
                c_kap, c_dln_kap_dlnRho, c_dln_kap_dlnT, ierr)
        else
           c_kap = 0
           c_dln_kap_dlnRho = 0
           c_dln_kap_dlnT = 0
        end if

        kap = alfa*w_kap + beta*c_kap
        dln_kap_dlnRho = (alfa*w_kap*w_dln_kap_dlnRho + beta*c_kap*c_dln_kap_dlnRho)/kap
        dln_kap_dlnT = (alfa*w_kap*w_dln_kap_dlnT + beta*c_kap*c_dln_kap_dlnT)/kap

      end subroutine cold_CO_kap_get_Type2


      integer function extras_startup(id, restart, ierr)
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         real :: log10_kap
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_startup = 0
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if

       end function extras_startup


      ! returns either keep_going, retry, backup, or terminate.
      integer function extras_check_model(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr, i
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_check_model = keep_going
         if (.false. .and. s% star_mass_h1 < 0.35d0) then
            ! stop when star hydrogen mass drops to specified level
            extras_check_model = terminate
            write(*, *) 'have reached desired hydrogen mass'
            return
         end if

         do i=1, s% nz, 1
                  if (s% mixing_type(i) == 1 .and. s% m(i) < 0.1*Msun) then
               s% D_mix(i) = 0d0
                  endif
                  if (s% q(i) < 1d-5) then
                     !s% D_mix(i) = 1d12                                                                                                        
            endif
         enddo


         ! if you want to check multiple conditions, it can be useful
         ! to set a different termination code depending on which
         ! condition was triggered.  MESA provides 9 customizeable
         ! termination codes, named t_xtra1 .. t_xtra9.  You can
         ! customize the messages that will be printed upon exit by
         ! setting the corresponding termination_code_str value.
         ! termination_code_str(t_xtra1) = 'my termination condition'

         ! by default, indicate where (in the code) MESA terminated
         if (extras_check_model == terminate) s% termination_code = t_extras_check_model
      end function extras_check_model


      integer function how_many_extra_history_columns(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_history_columns = 0
      end function how_many_extra_history_columns


      subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         integer :: k, i, k_conv
         real(dp) :: sigv, int, max_eps_nuc_entropy, power
         real, parameter, dimension(21) :: coeffs_n14 = &
           (/2.153390e+01, 0.000000e+00, -3.625040e+01, 0.000000e+00, 0.000000e+00, -5.000000e+00, -6.666670e-01, &
            1.968380e-01, -5.160340e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00, -1.500000e+00, &
            1.389950e+01, -1.096560e+01, -5.622700e+00, 0.000000e+00, 0.000000e+00, 0.000000e+00, -1.500000e+00/)
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return

         !note: do NOT add the extras names to history_columns.list
         ! the history_columns.list is only for the built-in log column options.
         ! it must not include the new column names you are adding here.

         do k = 1, s% nz
          if (s% mixing_type(k) .ne. 1) then
            k_conv = k-1
            exit
          endif
         enddo



           sigv = 0
           int = 0
           do k = 1, k_conv
            do i = 0, 2
              sigv = sigv + exp_cr(coeffs_n14(7*i + 1) + coeffs_n14(7*i + 2)/(s% T(k)*1d-9)&
               + coeffs_n14(7*i + 3)/(s% T(k)*1d-9)**(-1.0/3.0) + coeffs_n14(7*i + 4)* &
               (s% T(k)*1d-9)**(1.0/3.0) + coeffs_n14(7*i + 5)*(s% T(k)*1d-9)&
               + coeffs_n14(7*i + 6)*(s% T(k)*1d-9)**(5.0/3.0) + coeffs_n14(7*i + 7)* &
               safe_log_cr(s% T(k)*1d-9))
            enddo
            int = int + s% rho(k)*s% dm(k)*sigv
           enddo

         names(1) = 't_n14'

         vals(1) = 4*mp*(s% star_mass - s% conv_mx1_bot)*msol/(s% surface_he4*int)

         names(2) = 'max_eps_nuc_entropy'

         max_eps_nuc_entropy = 0d0
         power = 0d0
         do k = 1, s% nz
          if (s% eps_nuc(k) > power) then
            max_eps_nuc_entropy = s% entropy(k)
            power = s% eps_nuc(k)
          endif
        enddo

        vals(2) = max_eps_nuc_entropy

      end subroutine data_for_extra_history_columns


      integer function how_many_extra_profile_columns(id, id_extra)
         use star_def, only: star_info
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_profile_columns = 1
      end function how_many_extra_profile_columns


      subroutine data_for_extra_profile_columns(id, id_extra, n, nz, names, vals, ierr)
         use star_def, only: star_info, maxlen_profile_column_name
         use const_def, only: dp
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return

         !note: do NOT add the extra names to profile_columns.list
         ! the profile_columns.list is only for the built-in profile column options.
         ! it must not include the new column names you are adding here.

         ! here is an example for adding a profile column
         !if (n /= 1) stop 'data_for_extra_profile_columns'

         names(1) = 'kap_logR'
         do k = 1, nz
           vals(k,1) = safe_log10_cr(s% Rho(k)) - 3*safe_log10_cr(s% T(k)) + 18
         end do

      end subroutine data_for_extra_profile_columns


      ! returns either keep_going or terminate.
      ! note: cannot request retry or backup; extras_check_model can do that.
      integer function extras_finish_step(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_finish_step = keep_going
         call store_extra_info(s)

         ! to save a profile,
            ! s% need_to_save_profiles_now = .true.
         ! to update the star log,
            ! s% need_to_update_history_now = .true.

         ! see extras_check_model for information about custom termination codes
         ! by default, indicate where (in the code) MESA terminated
         if (extras_finish_step == terminate) s% termination_code = t_extras_finish_step
      end function extras_finish_step


      subroutine extras_after_evolve(id, id_extra, ierr)
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return

         ! clean up unnecessarily
         deallocate(kk_data)

      end subroutine extras_after_evolve


      ! routines for saving and restoring extra data so can do restarts

         ! put these defs at the top and delete from the following routines
         !integer, parameter :: extra_info_alloc = 1
         !integer, parameter :: extra_info_get = 2
         !integer, parameter :: extra_info_put = 3


      subroutine alloc_extra_info(s)
         integer, parameter :: extra_info_alloc = 1
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_alloc)
      end subroutine alloc_extra_info


      subroutine unpack_extra_info(s)
         integer, parameter :: extra_info_get = 2
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_get)
      end subroutine unpack_extra_info


      subroutine store_extra_info(s)
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_put)
      end subroutine store_extra_info


      subroutine move_extra_info(s,op)
         integer, parameter :: extra_info_alloc = 1
         integer, parameter :: extra_info_get = 2
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         integer, intent(in) :: op

         integer :: i, j, num_ints, num_dbls, ierr

         i = 0
         ! call move_int or move_flg
         num_ints = i

         i = 0
         ! call move_dbl

         num_dbls = i

         if (op /= extra_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return

         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if

         contains

         subroutine move_dbl(dbl)
            real(dp) :: dbl
            i = i+1
            select case (op)
            case (extra_info_get)
               dbl = s% extra_work(i)
            case (extra_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl

         subroutine move_int(int)
            integer :: int
            i = i+1
            select case (op)
            case (extra_info_get)
               int = s% extra_iwork(i)
            case (extra_info_put)
               s% extra_iwork(i) = int
            end select
         end subroutine move_int

         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (extra_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg

      end subroutine move_extra_info

      subroutine reduced_D_mix(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         integer :: i, k
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         ierr = 0

         !write(*,*) 'hello'

         do i=1, s% nz, 1
         !write(*,*) 'is this even doing anything?'
            if (s% mixing_type(i) == 1 .and. s% m(i) < 1.3*Msun) then
              s% D_mix(i) = 0d0
              !write(*,*) 'its working'
            endif
            if (s% q(i) < 1d-5) then
               !s% D_mix(i) = 1d12                                                                                                        
            endif
         enddo

      end subroutine reduced_D_mix


      end module run_star_extras
